/* This software was created by Hewlett-Packard, Co for the purposes of
 * demonstrating use of input methods in an internationalized Motif
 * application.  This software may be used, duplicated or modified
 * at will.  Hewlett-Packard, Co. provides this software "as is" without
 * warranty of any kind.
 */

#include "xmim.h"

void DA_Expose();
void DA_Input();
void DA_Destroy();

static XtResource resources [] = {
   {XmNfontList, XmCFontList, XmRFontList, sizeof(XmFontList),
    XtOffset(DrawDataPtr, fontlist), XmRString, "fixed"},
};

/* Global variables needed by the application */

static char initial_string[512]; 

nl_catd my_fd;          /* Message catalog file descriptor */

Widget Shell1, status;
Widget DrawArea;
Widget RowColumn1;


Boolean have_font_struct = False;
Boolean have_font_set = False;
XFontStruct *font;
GC my_gc = NULL;
XmFontContext context;

/* The function to Get the fontset out of fontlist for drawing. */

XFontSet FontList2FontSet( XmFontList fontlist)
{

XmFontListEntry next_entry;
XmFontType type_return = XmFONT_IS_FONT;
char* font_tag;
XFontSet fontset;
XFontSet first_fontset;
Boolean have_font_set = False;

if ( !XmFontListInitFontContext(&context, fontlist)) {
    XtWarning("fl2fs: can't create fontlist context...");
    exit(0);
}

while ((next_entry = XmFontListNextEntry(context)) != NULL) {

    fontset = (XFontSet) XmFontListEntryGetFont(next_entry, &type_return);
    if (type_return == XmFONT_IS_FONTSET ) {

        font_tag = XmFontListEntryGetTag(next_entry);

        if (!strcmp(XmFONTLIST_DEFAULT_TAG, font_tag)) {
            return fontset;
        }
        /* Remember the 1st fontset, just in case... */
        if (!have_font_set) {
            first_fontset = fontset;
            have_font_set = True;
        }
    }
}
if (have_font_set)
    return first_fontset;

return (XFontSet)NULL;
}


/* extern Boolean XtIsManaged(); */

void myQuit()
{
     printf(catgets(my_fd, ERROR_SET, 1,
	    "Begin exiting ... please standby ... \n"));
     catclose(my_fd);
     fflush(stdout);
     exit(0);
}

void 
#ifdef _NO_PROTO
InitializeDrawStruct(widget, data_ptr)
	DrawData *data_ptr;
	Widget widget;
#else
InitializeDrawStruct(
	Widget widget,
	DrawData *data_ptr )
#endif /* _NO_PROTO */
{
   int ascent = 0, descent = 0;
   XFontSetExtents *fs_extents;
   
   data_ptr->widget = widget;

  /* if we have one byte per character, store the data in the char*
   * portion of the structure; initialize the wchar_t portion to NULL.
   */

   if (MB_CUR_MAX == 1){
      strcpy(data_ptr->lines[0].text, initial_string);
      data_ptr->lines[0].num_chars = strlen(data_ptr->lines[0].text);
      data_ptr->lines[0].wc_text[0] = (wchar_t) NULL;
   } else {
  /* we need wchar_t to store one character per storage element (eases
   * character manipulation, such as processing backspace.  Use the
   * ANSI-C function mbstowcs to convert and store the data. */
      data_ptr->lines[0].num_chars = mbstowcs(data_ptr->lines[0].wc_text,
                                              initial_string, MAXLINESIZE);
     /* if an error occurs in conversion, the initial value must be bad.
      * so we'll just ignore the initial value and initialize to the
      * empty string, terminated with a wide character NULL.  */
      if (data_ptr->lines[0].num_chars < 0) {
         data_ptr->lines[0].num_chars = 0;
         data_ptr->lines[0].wc_text[0] = (wchar_t) NULL;
      }
      data_ptr->lines[0].text[0] = (char) NULL;
   }
  /* For this example, we'll assume a single line in the initial text.
   * If this was an actual application, we would check and count the number
   * of lines, filling out lines in the structure as appropriate.
   */
   data_ptr->num_lines = 1;
   data_ptr->active_line = 0;
   data_ptr->fontset = NULL;
   data_ptr->font = NULL;

  /* Call FontList2FontSet() to get the fontset. */

   data_ptr->fontset = FontList2FontSet(data_ptr->fontlist); 

  /* determine metrics of the font[set] to be used */
   if (data_ptr->fontset){
       fs_extents = XExtentsOfFontSet(data_ptr->fontset);

      /* Though MIT says to use max_logical_extents, many of the fonts
       * supplied with R4 and R5 have ink that extends beyond the bounding
       * box, making *DrawImage* difficult to use.  So we'll use
       * max_ink_extents to avoid this problem. */

      /* max_ink_extent.y is number of pixels from origin to top of
       * rectangle (i.e. y is negative) */

       ascent = -fs_extents->max_ink_extent.y;
       descent = fs_extents->max_ink_extent.height +
                                               fs_extents->max_ink_extent.y;
   } else if (data_ptr->font){
       ascent = data_ptr->font->max_bounds.ascent;
       descent = data_ptr->font->max_bounds.descent;
   }
   data_ptr->ascent = ascent;
   data_ptr->descent = descent;
   data_ptr->line_height = ascent + descent;

  /* arbitrarily pick an x,y location for the start of the first line */
   data_ptr->lines[0].location.x = 20;
   data_ptr->lines[0].location.y = 5 + data_ptr->line_height;
}

void
#ifdef _NO_PROTO
InitializeIM(widget, data)
        Widget widget;
        DrawData *data;
#else
InitializeIM(
	Widget widget,
        DrawData *data)
#endif /* _NO_PROTO */
{
   Arg args[6];  /* To set initial values to input method */
   Cardinal n;
   XPoint xmim_point;
   XGCValues values;
   Pixmap bg_pixmap;
   int active_line = data->active_line;
   int data_pixel_length = 0;

  /* start the input method */
   XmImRegister(widget, (unsigned int)NULL);

  /* determine background, foreground, and background pixmap to be
   * passed to the input method.
   */
   n = 0;
   XtSetArg (args[n], XtNforeground, &values.foreground); n++;
   XtSetArg (args[n], XtNbackground, &values.background); n++;
   XtSetArg (args[n], XtNbackgroundPixmap, &bg_pixmap); n++;
   XtGetValues (widget, args, n);

  /* provide the x,y location of where the next character will be displayed,
   * so if the input method needs to pop up a window (i.e. for over the spot
   * input method), the window will be where the user is looking.
   */

   if (data->fontset){
      if (MB_CUR_MAX == 1)
         data_pixel_length = XmbTextEscapement(data->fontset,
                                               data->lines[active_line].text,
                                            data->lines[active_line].num_chars);
      else
         data_pixel_length = XwcTextEscapement(data->fontset,
                                               data->lines[active_line].wc_text,
                                            data->lines[active_line].num_chars);
   } else if (data->font){
      data_pixel_length = XTextWidth(data->font, data->lines[active_line].text,
                                     data->lines[active_line].num_chars);
   }

   xmim_point.x = data->lines[active_line].location.x + data_pixel_length;
   xmim_point.y = data->lines[active_line].location.y;

   n = 0;
   XtSetArg(args[n], XmNfontList, data->fontlist); n++;
   XtSetArg(args[n], XmNbackground, values.background); n++;
   XtSetArg(args[n], XmNforeground, values.foreground); n++;
   XtSetArg(args[n], XmNbackgroundPixmap, bg_pixmap);n++;
   XtSetArg(args[n], XmNspotLocation, &xmim_point); n++;
   XmImSetValues(widget, args, n);
}

void
#ifdef _NO_PROTO
IncrementActiveLine(data, x, y)
        DrawData *data;
	int x;
	int y;
#else
IncrementActiveLine(
        DrawData *data,
	int x, 
	int y )
#endif /* _NO_PROTO */
{
   int active_line;

   /* If this was a real application, we would need to do some fancy
    * calculation of line location, and allow the user to easily change
    * the active line.  However, for a simple example of input methods
    * in drawing areas, we'll just increment the current active line
    * and set the x,y and num_chars for a new line.  If we hit MAXLINES,
    * leave it at the last line.
    */

   if (data->num_lines < MAXLINES) {
      data->active_line++;
      active_line = data->active_line;
      data->num_lines++;
      data->lines[active_line].location.x =
                                    data->lines[active_line - 1].location.x;
      data->lines[active_line].location.y =
                                   data->lines[active_line - 1].location.y + 
				   data->line_height;
      data->lines[active_line].num_chars = 0;
   } else 
      data->active_line = MAXLINES - 1;
}


 /* The following routines is used to erase a line allowing the udpated
  * line to be re-displayed.  It is necessary to erase the entire line
  * to correctly display the result of backspace.  If this was a real
  * application, we would try to be tricky and calculate the minimum
  * amount of real estate to be touched.
  */

void
#ifdef _NO_PROTO
EraseLine(data, line_number)
	DrawData *data;
	int line_number;
#else
EraseLine(
	DrawData *data,
	int line_number)
#endif /* _NO_PROTO */
{
   int n = 0;
   int x_seg_len = 0;
   XGCValues values;
   unsigned long valuemask;
   Arg args[2];

  /* create the GC for drawing text in the drawing area */
   valuemask = GCForeground | GCBackground;

  /* swap the foreground/background colors in the GC to do the erasure. */
   n = 0;
   XtSetArg (args[n], XtNforeground, &values.background); n++;
   XtSetArg (args[n], XtNbackground, &values.foreground); n++;
   XtGetValues (data->widget, args, n);

   XChangeGC(XtDisplay(data->widget), my_gc, valuemask, &values);

   if (data->fontset && MB_CUR_MAX == 1)
      x_seg_len = XmbTextEscapement(data->fontset, 
				    data->lines[line_number].text, 
				    data->lines[line_number].num_chars);
   else if (data->fontset)
      x_seg_len = XwcTextEscapement(data->fontset, 
				    data->lines[line_number].wc_text, 
				    data->lines[line_number].num_chars);
   else if (data->font)
     /* can't reaonably handle MB_CUR_MAX >1 with font structs, so don't try */
      x_seg_len = XTextWidth(data->font, data->lines[line_number].text, 
			       data->lines[line_number].num_chars);

  /* We'll add a few pixels to the length to cover erasure of italicized
   * text, which tends to run *past* the calculated length.
   */
   if (x_seg_len)
      XFillRectangle(XtDisplay(data->widget), XtWindow(data->widget), 
		     my_gc, data->lines[line_number].location.x,
		     data->lines[line_number].location.y - data->ascent,
		     x_seg_len + 5, data->line_height);

  /* restore the GC */
   n = 0;
   XtSetArg (args[n], XtNforeground, &values.foreground); n++;
   XtSetArg (args[n], XtNbackground, &values.background); n++;
   XtGetValues (data->widget, args, n);

   XChangeGC(XtDisplay(data->widget), my_gc, valuemask, &values);

}

void
#ifdef _NO_PROTO
DA_Destroy (w, data, cb)
        Widget w;
        DrawData *data;
        XmDrawingAreaCallbackStruct *cb;
#else
DA_Destroy (
        Widget w,
        DrawData *data,
        XmDrawingAreaCallbackStruct *cb )
#endif /* _NO_PROTO */
{

   XmImUnregister(data->widget);

}


void
#ifdef _NO_PROTO
DA_Input (w, data, cb)
        Widget w;
        DrawData *data;
        XmDrawingAreaCallbackStruct *cb;
#else
DA_Input (
        Widget w,
        DrawData *data,
        XmDrawingAreaCallbackStruct *cb )
#endif /* _NO_PROTO */
{
   Status status_return;
   int insert_length=0;
   char insert_string[TEXT_MAX_INSERT_SIZE];
   wchar_t wc_insert_string[TEXT_MAX_INSERT_SIZE];
   int active_line = data->active_line;
   int position = 0;
   int num_chars = 0;
   int cur_end = 0;
   XPoint xmim_point;

   insert_string[0] = NULL;

   switch (cb->event->type) {
      case KeyPress: 

        /* Get any characters entered from the keyboard; the input method
         * will set insert_length > 0 when there is data.  */

         insert_length = XmImMbLookupString(w, (XKeyEvent *) cb->event,
                                            insert_string,
                                            TEXT_MAX_INSERT_SIZE,
                                            (KeySym *) NULL, &status_return);

        /* process new_line */
	 if (insert_length == 1 && insert_string[0] == '
')
	    IncrementActiveLine(data, data->lines[active_line].location.x,
			  data->lines[active_line].location.y);
        /* process backspace */
	 else if (insert_length ==1 && insert_string[0] == ''){
	    if (data->lines[active_line].num_chars > 0){
	       EraseLine(data, active_line);
	       position = --data->lines[active_line].num_chars;
	       if (MB_CUR_MAX == 1)
	          data->lines[active_line].text[position] = NULL;
	       else 
	          data->lines[active_line].wc_text[position] = NULL;
	    } else
	       if (active_line > 0)
		  data->active_line = --active_line;
	 } else {

        /* if there's room, append it; otherwise, replace it */

	    if (MB_CUR_MAX == 1) {
               if (insert_length + data->lines[active_line].num_chars
							        <= MAXLINESIZE){
                  (void) strncat(data->lines[active_line].text, insert_string,
			         insert_length);
                  data->lines[active_line].num_chars += insert_length;
               } else {
                  (void) strncpy(data->lines[active_line].text, insert_string,
			        insert_length);
	          data->lines[active_line].text[insert_length] = NULL;
                  data->lines[active_line].num_chars = insert_length;
	       }
	    } else {
               insert_string[insert_length] = '\0'; /* null terminate it */
               num_chars = mbstowcs(wc_insert_string, insert_string,
                                    insert_length + 1);
               if (num_chars + data->lines[active_line].num_chars
							        <= MAXLINESIZE){
                  cur_end = data->lines[active_line].num_chars;
                  (void) memcpy(&data->lines[active_line].wc_text[cur_end],
                                wc_insert_string, num_chars * sizeof(wchar_t));
                  data->lines[active_line].num_chars += num_chars;
               } else {
                  (void) memcpy(data->lines[active_line].wc_text,
                                wc_insert_string, num_chars * sizeof(wchar_t));
                  data->lines[active_line].num_chars = num_chars;
                  data->lines[active_line].wc_text[num_chars] = (wchar_t) NULL;
               }
	    }
	 }

        /* force the modified string to be displayed. If this was a real
         * application, we would be much more efficient about this. */

         DA_Expose(w, data, cb);

        /* tell the IM the new location of the input location */

         xmim_point.y = data->lines[active_line].location.y;
         if (data->fontset){
            if (MB_CUR_MAX == 1)
               xmim_point.x = data->lines[active_line].location.x + 
                              XmbTextEscapement(data->fontset,
                                                data->lines[active_line].text,
                                            data->lines[active_line].num_chars);
            else
               xmim_point.x = data->lines[active_line].location.x +
                              XwcTextEscapement(data->fontset,
                                               data->lines[active_line].wc_text,
                                            data->lines[active_line].num_chars);
         } else if (data->font){
            xmim_point.x = data->lines[active_line].location.x + 
                              XTextWidth(data->font, 
                                         data->lines[active_line].text,
                                         data->lines[active_line].num_chars);
         }
         XmImVaSetValues(w, XmNspotLocation, &xmim_point, NULL);
         break;
   }
}


void
#ifdef _NO_PROTO
CreateGC (data)
	DrawData *data;
#else
CreateGC (
	DrawData *data)

#endif /* _NO_PROTO */
{
   Arg args[20];
   int n;
   XGCValues values;
   unsigned long valuemask;

  /* create the GC for drawing text in the drawing area */
   valuemask = GCForeground | GCBackground;

   n = 0;
   XtSetArg (args[n], XtNforeground, &values.foreground); n++;
   XtSetArg (args[n], XtNbackground, &values.background); n++;
   XtGetValues (data->widget, args, n);

   if (!data->fontset) { /* We don't have a fontset, so add fontstruct to GC */
      valuemask |= GCFont;
      values.font = data->font->fid;
   }

   my_gc = XCreateGC(XtDisplay(DrawArea), XtWindow(DrawArea), valuemask, 
                     &values);
}

void 
#ifdef _NO_PROTO
DA_Expose (w, data, cb)
	Widget w;
	DrawData *data;
	XmDrawingAreaCallbackStruct *cb;
#else
DA_Expose (
	Widget w,
	DrawData *data,
	XmDrawingAreaCallbackStruct *cb )
#endif /* _NO_PROTO */
{
   int line;
   MyLineRec cur_line;


   if (!XtIsRealized((Widget)DrawArea)) return;

   if (!my_gc) CreateGC(data);

   if (data->fontset) {
      for (line = data->num_lines - 1; line >= 0; line--) {
	 cur_line = data->lines[line];
         if (MB_CUR_MAX == 1)
            XmbDrawString (XtDisplay(DrawArea), XtWindow(DrawArea),
                           data->fontset, my_gc, 
                           (int)cur_line.location.x, (int)cur_line.location.y, 
                           cur_line.text, cur_line.num_chars);
         else
            XwcDrawString (XtDisplay(DrawArea), XtWindow(DrawArea),
                           data->fontset, my_gc,
                           (int)cur_line.location.x, (int)cur_line.location.y,
                           cur_line.wc_text, cur_line.num_chars);
      }
   } else {
      for (line = data->num_lines - 1; line >= 0; line--) {
	 cur_line = data->lines[line];
         XDrawString (XtDisplay(DrawArea), XtWindow(DrawArea),
                      my_gc, (int)cur_line.location.x, (int)cur_line.location.y,
                      cur_line.text, cur_line.num_chars);
      }
   }
}


/* Tell the input method when we gain or lose focus.  This allows it
 * to alter its visuals to indicate its focus status.  In addition, because
 * in some models/applications a focus change can accompany an input location
 * change (for example, text widget in explicit focus mode), we tell the IM
 * the location of the input on focus change.  This isn't really needed in
 * this sample program, but in a real application (that allowed the user
 * to specify where data was to be drawn/input) this *would* be necessary. */ 

void
#ifdef _NO_PROTO
HandleFocusChange (w, data, event)
        Widget w;
        DrawData *data;
        XEvent *event;
#else
HandleFocusChange (
        Widget w,
        DrawData *data,
        XEvent *event )
#endif /* _NO_PROTO */
{
   Arg args[20];
   int n = 0;
   XPoint xmim_point;
   int active_line = data->active_line;
   int data_pixel_length = 0;

   if (data->fontset){
      if (MB_CUR_MAX == 1)
         data_pixel_length = XmbTextEscapement(data->fontset,
                                               data->lines[active_line].text, 
                                            data->lines[active_line].num_chars);
      else
         data_pixel_length = XwcTextEscapement(data->fontset,
                                               data->lines[active_line].wc_text,
                                            data->lines[active_line].num_chars);
   } else if (data->font){
      data_pixel_length = XTextWidth(data->font, data->lines[active_line].text,
                                     data->lines[active_line].num_chars);
   }

   xmim_point.x = data->lines[active_line].location.x + data_pixel_length;
   xmim_point.y = data->lines[active_line].location.y;

   if (event->type = FocusIn){
      XtSetArg(args[n], XmNspotLocation, &xmim_point); n++;
      XmImSetFocusValues(data->widget, args, n);
   } else if (event->type = FocusOut)
      XmImUnsetFocus(w);
}

void main (argc, argv)
int argc;
char **argv;
{
   Widget BulletinBoard;
   Display        *display;
   Arg args[20];
   int n;
   static char default_text[] = "This is the starting string";
   DrawData data;

   char *dummy;

   my_fd = catopen("drawIM",0);

   dummy=catgets(my_fd, VALUE_SET, 1, default_text);
   strcpy(initial_string, dummy);

   (void)XtSetLanguageProc(NULL, NULL, NULL);


   signal (SIGINT, myQuit);

   /*  initialize toolkit  */
   XtToolkitInitialize();
   app_context = XtCreateApplicationContext();
   display = XtOpenDisplay(app_context, NULL, argv[0], "XMclient", NULL, 0,
                            &argc, argv);
   if (!display) {
       printf(catgets(my_fd, ERROR_SET, 3, "Unable to open display\n"));
       exit(0);
   }

   /* create message shell */
   n = 0;
   XtSetArg (args[n], XmNallowShellResize, True);        n++;
   Shell1 = XtAppCreateShell(argv[0], NULL, applicationShellWidgetClass,
                            display, args, n);

   XtGetApplicationResources(Shell1, &data, resources, XtNumber(resources),
                             NULL, 0);

   n = 0;
   BulletinBoard = XmCreateBulletinBoard (Shell1, "BulletinBoard", args, n);
   XtManageChild (BulletinBoard);

   n = 0;
   XtSetArg (args[n], XmNresizePolicy, XmRESIZE_NONE);        n++;
   XtSetArg (args[n], XmNheight, 500);        n++;
   XtSetArg (args[n], XmNwidth, 800);        n++;
   DrawArea = XmCreateDrawingArea (BulletinBoard, "DrawArea", args, n);
   XtAddCallback (DrawArea, XmNexposeCallback, (XtCallbackProc) DA_Expose, 
		  (XtPointer)&data);
   XtAddCallback (DrawArea, XmNinputCallback, (XtCallbackProc) DA_Input, 
		  (XtPointer)&data);
   XtAddCallback (DrawArea, XmNdestroyCallback, (XtCallbackProc) DA_Destroy, 
		  (XtPointer)&data);
   XtManageChild (DrawArea);

   XtAddEventHandler(BulletinBoard, FocusChangeMask, False, 
		     (XtEventHandler)HandleFocusChange, (XtPointer)&data);

   InitializeDrawStruct(DrawArea, &data);
   InitializeIM(DrawArea, &data);

   XtRealizeWidget(Shell1);
   XtAppMainLoop(app_context);
}
